import unittest

from nebulento import IntentContainer, MatchStrategy


class TestIntentContainer(unittest.TestCase):
    def test_syntax(self):
        container = IntentContainer()
        container.add_intent('hello', ["(hello|hi|hey) world"])
        self.assertEqual(set(container.registered_intents["hello"]),
                         set(['hello world', 'hi world', 'hey world']))

        container = IntentContainer()
        container.add_intent('hello', ["hello (world|)"])
        self.assertEqual(set(container.registered_intents["hello"]),
                         set(['hello world', 'hello']))

        container.add_intent('hey', ["hey [world]"])
        self.assertEqual(set(container.registered_intents["hey"]),
                         set(['hey world', 'hey']))

        container.add_intent('hi', ["hi [{person}|people]"])
        self.assertEqual(set(container.registered_intents["hi"]),
                         set(['hi {person}', 'hi people', 'hi']))

    # test intent parsing
    def test_intents(self):
        container = IntentContainer()
        container.add_intent('hello', [
            'hello', 'hi', 'how are you', "what's up"
        ])
        container.add_intent('buy', [
            'buy {item}', 'purchase {item}', 'get {item}', 'get {item} for me'
        ])
        container.add_entity('item', [
            'milk', 'cheese'
        ])
        self.assertEqual(container.calc_intent('hello'),
                         {'best_match': 'hello',
                          'conf': 1.0,
                          'entities': {},
                          'match_strategy': 'DAMERAU_LEVENSHTEIN_SIMILARITY',
                          'name': 'hello',
                          'utterance': 'hello',
                          'utterance_consumed': 'hello',
                          'utterance_remainder': ''})

        self.assertEqual(container.calc_intent('buy milk'),
                         {'best_match': 'buy {item}',
                          'conf': 0.625,
                          'entities': {'item': ['milk']},
                          'match_strategy': 'DAMERAU_LEVENSHTEIN_SIMILARITY',
                          'name': 'buy',
                          'utterance': 'buy milk',
                          'utterance_consumed': 'buy milk',
                          'utterance_remainder': ''})
        self.assertEqual(container.calc_intent('buy beer'),
                         {'best_match': 'buy {item}',
                          'conf': 0.5,
                          'entities': {},
                          'match_strategy': 'DAMERAU_LEVENSHTEIN_SIMILARITY',
                          'name': 'buy',
                          'utterance': 'buy beer',
                          'utterance_consumed': 'buy',
                          'utterance_remainder': 'beer'}
                         )

    def test_case(self):
        container = IntentContainer()
        container.add_intent('test', ['Testing cAPitalizAtion'])
        self.assertEqual(
            container.calc_intent('Testing cAPitalizAtion')['conf'], 1.0)
        self.assertEqual(
            container.calc_intent('teStiNg CapitalIzation')['conf'], 1.0)

        container = IntentContainer(ignore_case=False)
        container.add_intent('test', ['Testing cAPitalizAtion'])
        self.assertEqual(
            container.calc_intent('teStiNg CapitalIzation')['conf'],
            0.6363636363636364)

    def test_entities(self):
        container = IntentContainer()
        container.add_intent('test3', ['I see {thing} (in|on) {place}'])
        self.assertEqual(
            container.calc_intent('I see a bin in there'),
            {'best_match': 'i see {thing} in {place}',
             'conf': 0.5416666666666667,
             'entities': {},
             'match_strategy': 'DAMERAU_LEVENSHTEIN_SIMILARITY',
             'name': 'test3',
             'utterance': 'i see a bin in there',
             'utterance_consumed': 'i see in',
             'utterance_remainder': 'a bin there'}
        )
        container.add_entity("place", ["floor", "table"])
        self.assertEqual(
            container.calc_intent('I see trash in the floor'),
            {'best_match': 'i see {thing} in {place}',
             'conf': 0.53125,
             'entities': {'place': ['floor']},
             'match_strategy': 'DAMERAU_LEVENSHTEIN_SIMILARITY',
             'name': 'test3',
             'utterance': 'i see trash in the floor',
             'utterance_consumed': 'i see in floor',
             'utterance_remainder': 'trash the'}
        )
        container.add_entity("thing", ["food"])
        self.assertEqual(
            container.calc_intent('I see food in the table'),
            {'best_match': 'i see {thing} in {place}',
             'conf': 0.6484375,
             'entities': {'place': ['table'], 'thing': ['food']},
             'match_strategy': 'DAMERAU_LEVENSHTEIN_SIMILARITY',
             'name': 'test3',
             'utterance': 'i see food in the table',
             'utterance_consumed': 'i see in table food',
             'utterance_remainder': 'the'}
        )

    # test match strategies
    def test_token_set(self):
        container = IntentContainer(
            fuzzy_strategy=MatchStrategy.TOKEN_SET_RATIO)
        container.add_intent('buy', [
            'buy {item}', 'purchase {item}', 'get {item}', 'get {item} for me'
        ])
        container.add_entity('item', [
            'milk', 'cheese'
        ])
        self.assertEqual(container.calc_intent('buy milk'),
                         {'best_match': 'buy {item}',
                          'conf': 0.6666666666666667,
                          'entities': {'item': ['milk']},
                          'match_strategy': 'TOKEN_SET_RATIO',
                          'name': 'buy',
                          'utterance': 'buy milk',
                          'utterance_consumed': 'buy milk',
                          'utterance_remainder': ''})
        self.assertEqual(container.calc_intent('buy beer'),
                         {'best_match': 'buy {item}',
                          'conf': 0.5555555555555556,
                          'entities': {},
                          'match_strategy': 'TOKEN_SET_RATIO',
                          'name': 'buy',
                          'utterance': 'buy beer',
                          'utterance_consumed': 'buy',
                          'utterance_remainder': 'beer'})

    def test_token_sort(self):
        container = IntentContainer(
            fuzzy_strategy=MatchStrategy.TOKEN_SORT_RATIO)
        container.add_intent('buy', [
            'buy {item}', 'purchase {item}', 'get {item}', 'get {item} for me'
        ])
        container.add_entity('item', [
            'milk', 'cheese'
        ])
        self.assertEqual(container.calc_intent('buy milk'),
                         {'best_match': 'buy {item}',
                          'conf': 0.6666666666666667,
                          'entities': {'item': ['milk']},
                          'match_strategy': 'TOKEN_SORT_RATIO',
                          'name': 'buy',
                          'utterance': 'buy milk',
                          'utterance_consumed': 'buy milk',
                          'utterance_remainder': ''})
        self.assertEqual(container.calc_intent('buy beer'),
                         {'best_match': 'buy {item}',
                          'conf': 0.33333333333333337,
                          'entities': {},
                          'match_strategy': 'TOKEN_SORT_RATIO',
                          'name': 'buy',
                          'utterance': 'buy beer',
                          'utterance_consumed': 'buy',
                          'utterance_remainder': 'beer'})

    def test_partial_token_set(self):
        container = IntentContainer(
            fuzzy_strategy=MatchStrategy.PARTIAL_TOKEN_SET_RATIO)
        container.add_intent('buy', [
            'buy {item}', 'purchase {item}', 'get {item}', 'get {item} for me'
        ])
        container.add_entity('item', [
            'milk', 'cheese'
        ])
        self.assertEqual(container.calc_intent('buy milk'),
                         {'best_match': 'buy {item}',
                          'conf': 1,
                          'entities': {'item': ['milk']},
                          'match_strategy': 'PARTIAL_TOKEN_SET_RATIO',
                          'name': 'buy',
                          'utterance': 'buy milk',
                          'utterance_consumed': 'buy milk',
                          'utterance_remainder': ''})
        self.assertEqual(container.calc_intent('buy beer'),
                         {'best_match': 'buy {item}',
                          'conf': 1.0,
                          'entities': {},
                          'match_strategy': 'PARTIAL_TOKEN_SET_RATIO',
                          'name': 'buy',
                          'utterance': 'buy beer',
                          'utterance_consumed': 'buy',
                          'utterance_remainder': 'beer'})

    def test_partial_token_sort(self):
        container = IntentContainer(
            fuzzy_strategy=MatchStrategy.PARTIAL_TOKEN_SORT_RATIO)
        container.add_intent('buy', [
            'buy {item}', 'purchase {item}', 'get {item}', 'get {item} for me'
        ])
        container.add_entity('item', [
            'milk', 'cheese'
        ])
        self.assertEqual(container.calc_intent('buy milk'),
                         {'best_match': 'buy {item}',
                          'conf': 0.7857142857142857,
                          'entities': {'item': ['milk']},
                          'match_strategy': 'PARTIAL_TOKEN_SORT_RATIO',
                          'name': 'buy',
                          'utterance': 'buy milk',
                          'utterance_consumed': 'buy milk',
                          'utterance_remainder': ''})
        self.assertEqual(container.calc_intent('buy beer'),
                         {'best_match': 'buy {item}',
                          'conf': 0.5454545454545454,
                          'entities': {},
                          'match_strategy': 'PARTIAL_TOKEN_SORT_RATIO',
                          'name': 'buy',
                          'utterance': 'buy beer',
                          'utterance_consumed': 'buy',
                          'utterance_remainder': 'beer'})



import itertools
import re
from typing import List, Dict


def expand_template(template: str) -> List[str]:
    def expand_optional(text):
        """Replace [optional] with two options: one with and one without."""
        return re.sub(r"\[([^\[\]]+)\]", lambda m: f"({m.group(1)}|)", text)

    def expand_alternatives(text):
        """Expand (alternative|choices) into a list of choices."""
        parts = []
        for segment in re.split(r"(\([^\(\)]+\))", text):
            if segment.startswith("(") and segment.endswith(")"):
                options = segment[1:-1].split("|")
                parts.append(options)
            else:
                parts.append([segment])
        return itertools.product(*parts)

    def fully_expand(texts):
        """Iteratively expand alternatives until all possibilities are covered."""
        result = set(texts)
        while True:
            expanded = set()
            for text in result:
                options = list(expand_alternatives(text))
                expanded.update(["".join(option).strip() for option in options])
            if expanded == result:  # No new expansions found
                break
            result = expanded
        return sorted(result)  # Return a sorted list for consistency

    # Expand optional items first
    template = expand_optional(template)

    # Fully expand all combinations of alternatives
    return fully_expand([template])


def expand_slots(template: str, slots: Dict[str, List[str]]) -> List[str]:
    """Expand a template by first expanding alternatives and optional components,
    then substituting slot placeholders with their corresponding options.

    Args:
        template (str): The input string template to expand.
        slots (dict): A dictionary where keys are slot names and values are lists of possible replacements.

    Returns:
        list[str]: A list of all expanded combinations.
    """
    # Expand alternatives and optional components
    base_expansions = expand_template(template)

    # Process slots
    all_sentences = []
    for sentence in base_expansions:
        matches = re.findall(r"\{([^\{\}]+)\}", sentence)
        if matches:
            # Create all combinations for slots in the sentence
            slot_options = [slots.get(match, [f"{{{match}}}"]) for match in matches]
            for combination in itertools.product(*slot_options):
                filled_sentence = sentence
                for slot, replacement in zip(matches, combination):
                    filled_sentence = filled_sentence.replace(f"{{{slot}}}", replacement)
                all_sentences.append(filled_sentence)
        else:
            # No slots to expand
            all_sentences.append(sentence)

    return all_sentences



import logging
from nebulento.fuzz import MatchStrategy, match_one
from nebulento.bracket_expansion import expand_template, expand_slots
import quebra_frases

LOG = logging.getLogger('nebulento')


class IntentContainer:
    def __init__(self, fuzzy_strategy=MatchStrategy.DAMERAU_LEVENSHTEIN_SIMILARITY,
                 ignore_case=True):
        self.fuzzy_strategy = fuzzy_strategy
        self.ignore_case = ignore_case
        self.registered_intents = {}
        self.registered_entities = {}

    def match_entities(self, sentence):
        if self.ignore_case:
            sentence = sentence.lower()
        matches = {}
        for entity, samples in self.registered_entities.items():
            chunked = quebra_frases.chunk(sentence, samples)
            matches[entity] = [s for s in samples if s in chunked]
        return matches

    def match_fuzzy(self, sentence):
        if self.ignore_case:
            sentence = sentence.lower()
        entities = self.match_entities(sentence)
        for intent, samples in self.registered_intents.items():
            samples = self.registered_intents[intent]

            sent, score = match_one(sentence, samples,
                                    strategy=self.fuzzy_strategy)
            remainder = [
                w for w in quebra_frases.word_tokenize(sentence)
                if w not in quebra_frases.word_tokenize(sent)]
            consumed = [
                w for w in quebra_frases.word_tokenize(sentence)
                if w in quebra_frases.word_tokenize(sent)]

            tagged_entities = {}
            for ent, v in entities.items():
                if v and any("{" + ent + "}" in s for s in samples):
                    score = 0.25 + score * 0.75
                    tagged_entities[ent] = v
                    consumed += [_ for _ in v if _ not in consumed]
                    remainder = [_ for _ in remainder if _ not in v]
            remainder = " ".join(remainder)
            consumed = " ".join(consumed)
            yield {"best_match": sent,
                   "conf": min(score, 1),
                   "entities": tagged_entities,
                   "match_strategy": self.fuzzy_strategy.name,
                   "utterance": sentence,
                   "utterance_remainder": remainder,
                   "utterance_consumed": consumed,
                   "name": intent}

    def add_intent(self, name, lines):
        expanded = []
        for l in lines:
            expanded += expand_template(l)
        if self.ignore_case:
            expanded = [l.lower() for l in expanded]
        self.registered_intents[name] = expanded

    def remove_intent(self, name):
        if name in self.registered_intents:
            del self.registered_intents[name]

    def add_entity(self, name, lines):
        expanded = []
        for l in lines:
            expanded += expand_template(l)
        if self.ignore_case:
            expanded = [l.lower() for l in expanded]
        self.registered_entities[name] = expanded

    def remove_entity(self, name):
        if name in self.registered_entities:
            del self.registered_entities[name]

    def calc_intents(self, query):
        for intent in self.match_fuzzy(query):
            yield intent

    def calc_intent(self, query):
        return max(
            self.calc_intents(query),
            key=lambda x: x["conf"],
            default={"best_match": None,
                     "conf": 0,
                     "match_strategy": self.fuzzy_strategy,
                     "utterance": query,
                     "name": None}
        )
